import matplotlib
matplotlib.use('WXAgg')
# template for Lab #2, Task #6
import numpy as np
import random
import matplotlib.pyplot as p
import lab2

def transmit(bits,npreamble=0,
			 npostamble=0,
			 samples_per_bit=8,
			 v0 = 0.0,
			 v1 = 1.0,
			 repeat = 1):

	samples = np.array([])

	for ntimes in range(repeat):
		
		for prefix in range(npreamble):
			samples = np.append(samples, [v0])
	
		for bit in bits:
			for s in range(samples_per_bit):
				if bit == 1:
					samples = np.append(samples, [v1])
				else:
					samples = np.append(samples, [v0])
				
		for postfix in range(npostamble):
			samples = np.append(samples, [v0])	
			
	return samples

def unit_sample_response(channel):
	
	samples = [1.0]+[0.0]*1000
	response = channel(samples)
	
	maxS = 0.0
	count = 0
	for s in range(1000):
		if response[s] > maxS:
			maxS = response[s]
		if response[s]*100 < maxS:
			count += 1
		else:
			count = 0
		if count >= 10:
			break
	
	return response[:s]

def deconvolver(y,h):
    
	#find start position in unit-sample-response
	startPos = 0
	for i in h:
		if i>0.001:
			break
		startPos+=1
	h = h[startPos:]
	k = len(y)
	
	#pad the unit-sample response
	if len(h)<k:
		h = np.append(h, [0 for zeros in range(len(y)-len(h))])
	
	#create the input estimator function
	w = []
	#deconvolve the output with the unit sample response
	for n in xrange(k):
		w.append(y[n])
		if n > 0:
			for r in xrange(1, n+1):
				w[n] -= (h[r]*w[n-r])
		w[n] /= h[0]
	
	w_numpy = np.array(w)
			
	return w_numpy

def test_channel(xbits, channel, samples_per_bit=8, noise=0.0):			
	x = transmit(xbits, 0, 100, samples_per_bit)
	
	h = unit_sample_response(channel)
	
	y = channel(x, noise)
	
	w = deconvolver(y, h)
	
	p.figure()
	plot_samples(411, x, x.shape[0], -.1, 1.1, 'Input')
	plot_samples(412, y, x.shape[0], -.1, 1.1, 'Output', 'g-')
	plot_samples(413, h, h.shape[0], -.1, 1.1, 'Unit Sample Response', 'b-')
	plot_samples(414, w, w.shape[0], -.1, 1.1, 'Deconvolved Result', 'r-')
	
def plot_samples(subplot, inp, size, ylo, yhi, title, color='b-'):
	
	xaxis = np.arange(size)
	p.subplot(subplot)
	p.plot(xaxis, inp, color)
	p.axis([0, size, ylo, yhi])
	p.ylabel('Voltage')
	p.grid(True)
	p.title(title)	

if __name__ == '__main__':
    
	xbits = [random.randint(0,1) for i in range(100)]
	test_channel(xbits, lab2.channel1, 10)
	test_channel(xbits, lab2.channel2, 10)
	
	p.show()